"""
connection to sht31 thermometer over REST API
data structure expected:
{
    "measurements": 1,
    "Temp(F) mean": 88.55024032959489,
    "Temp(C) std": 0.0,
    "Temp(F) std": 0.0,
    "Humidity(%RH) mean": 49.491111619745176,
    "Humidity(%RH) std": 0.0
}
"""
# built-in imports
import json
import os
import threading
import time
import traceback

# third party imports
import requests

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import sht31_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import utilities as util


class ThermostatClass(tc.ThermostatCommon):
    """SHT31 thermometer functions."""

    def __init__(self, zone, path=sht31_config.flask_folder.production,
                 verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat.
            path(str):  path on flask server, default = production
                        sht31_ip dict above must have correct IP address for
                        each zone.
            verbose(bool): debug flag.
        """
        # construct the superclass
        super().__init__()

        # set tstat type and debug flag
        self.thermostat_type = sht31_config.ALIAS
        self.verbose = verbose

        # zone configuration
        self.zone_name = int(zone)
        self.ip_address = self.get_target_zone_id(self.zone_name)
        self.path = path

        # URL and port configuration
        self.port = str(sht31_config.FLASK_PORT)  # Flask server port on host
        if (self.zone_name == sht31_config.UNIT_TEST_ZONE and
                self.path == sht31_config.flask_folder.production):
            self.path = sht31_config.flask_folder.unit_test
            self.unit_test_seed = "?seed=" + str(sht31_config.UNIT_TEST_SEED)
        else:
            # self.path = ""
            self.unit_test_seed = ""
        self.measurements = "?measurements=" + str(sht31_config.MEASUREMENTS)
        self.url = (sht31_config.FLASK_URL_PREFIX + self.ip_address + ":" +
                    self.port + self.path + self.measurements +
                    self.unit_test_seed)
        self.device_id = self.url
        self.retry_delay = 60  # delay before retrying a bad reading

        # if in unit test mode, spawn flask server with emulated data on client
        if self.zone_name == sht31_config.UNIT_TEST_ZONE:
            self.spawn_flask_server()

    def get_target_zone_id(self, zone=0):
        """
        Return the target zone ID (aka IP address for sht31) from the
        zone number provided.

        inputs:
            zone(int): specified zone number
        returns:
            (str):  IP address of target zone.
        """
        env_str = self.get_env_key(zone)
        ip_address = self.get_ip_address(env_str)
        return ip_address

    def get_env_key(self, zone_str):
        """
        Return env key for the zone specified.

        inputs:
            zone_str(str or int): zone number
        returns:
            (str): env var key
        """
        return 'SHT31_REMOTE_IP_ADDRESS' + '_' + str(zone_str)

    def get_ip_address(self, env_key):
        """
        Return IP address from env key and cache value in dict.

        inputs:
            env_key(str): env var key.
        returns:
            (str):  IP address
        """
        return env.get_env_variable(env_key)["value"]

    def spawn_flask_server(self):
        """
        Spawn a local flask server for unit testing.

        inputs: None
        returns:
        """
        # flask server used in unit test mode
        from thermostatsupervisor import sht31_flask_server as sht31_fs  # noqa E402, C0415
        # pylint: disable=import-outside-toplevel

        # setup flask runtime variables
        sht31_fs.uip = sht31_fs.UserInputs(
            [os.path.realpath(__file__),
             sht31_config.FLASK_DEBUG_MODE])

        # start flask server thread
        self.flask_server = threading.Thread(
            target=sht31_fs.app.run,
            args=('0.0.0.0', sht31_config.FLASK_PORT,
                  sht31_config.FLASK_DEBUG_MODE),
            kwargs=sht31_config.FLASK_KWARGS)
        self.flask_server.daemon = True  # make thread daemonic
        self.flask_server.start()
        util.log_msg(f"thread alive status={self.flask_server.is_alive()}",
                     mode=util.BOTH_LOG, func_name=1)
        util.log_msg("Flask server setup is complete",
                     mode=util.BOTH_LOG, func_name=1)

    def print_all_thermostat_metadata(self, zone, debug=False):
        """
        Print initial meta data queried from thermostat.

        inputs:
            zone(int): target zone
            debug(bool): debug flag
        returns:
            (dict): return data
        """
        return_data = self.exec_print_all_thermostat_metadata(
            self.get_all_metadata, [zone, debug])
        return return_data

    def get_all_metadata(self, zone, retry=True):
        """
        Get all the current thermostat metadata.

        inputs:
            zone(int): target zone
            retry(bool): if True will retry once.
        returns:
            (dict) results dict.
        """
        return self.get_metadata(zone, None, retry)

    def get_metadata(self, zone, parameter=None, retry=True):
        """
        Get current thermostat metadata.

        inputs:
            zone(str or int): (unused) target zone
            parameter(str): target parameter, if None will fetch all.
            retry(bool): if True will retry once.
        returns:
            (dict) empty dict.
        """
        try:
            response = requests.get(self.url, timeout=util.HTTP_TIMEOUT)
        except requests.exceptions.ConnectionError as ex:
            util.log_msg(
                f"FATAL ERROR: unable to connect to sht31 thermometer at url "
                f"'{self.url}'",
                mode=util.BOTH_LOG,
                func_name=1)
            raise ex
        try:
            if parameter is None:
                return response.json()
            else:
                return response.json()[parameter]
        except json.decoder.JSONDecodeError as ex:
            util.log_msg(traceback.format_exc(),
                         mode=util.BOTH_LOG,
                         func_name=1)
            if retry:
                util.log_msg(
                    f"waiting {self.retry_delay} seconds and retrying SHT31 "
                    f"measurement one time...",
                    mode=util.BOTH_LOG,
                    func_name=1)
                time.sleep(self.retry_delay)
                self.get_metadata(zone, parameter, retry=False)
            else:
                raise RuntimeError("FATAL ERROR: SHT31 server "
                                   "is not responding") from ex


class ThermostatZone(tc.ThermostatCommonZone):
    """SHT31 thermometer zone functions."""

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            Thermostat_obj(obj): associated Thermostat_obj
            verbose(bool): debug flag.
        """
        # construct the superclass
        super().__init__()

        # switch config for this thermostat
        # SHT31 is a monitor only, does not support heat/cool modes.
        self.system_switch_position[tc.ThermostatCommonZone.OFF_MODE] = 0

        # zone configuration
        self.verbose = verbose
        self.thermostat_type = sht31_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.url = Thermostat_obj.device_id
        self.zone_name = Thermostat_obj.zone_name
        self.zone_name = self.get_zone_name(self.zone_name)

        # runtime defaults
        self.poll_time_sec = 1 * 60  # default to 1 minute
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

        self.tempfield = sht31_config.API_TEMPF_MEAN  # must match flask API
        self.humidityfield = sht31_config.API_HUMIDITY_MEAN  # must match API
        self.retry_delay = Thermostat_obj.retry_delay

    def get_zone_name(self, zone) -> str:  # used
        """
        Return zone name.

        inputs:
            zone(int): zone number
        returns:
            (str): zone name
        """
        return sht31_config.metadata[zone]["zone_name"]

    def get_metadata(self, parameter=None, retry=True):
        """
        Get the current thermostat metadata settings.

        inputs:
          parameter(str): target parameter, None = all settings
          retry(bool): if True, will retry on Exception
        returns:
          dict if parameter=None
          str if parameter != None
        """
        try:
            response = requests.get(self.url, timeout=util.HTTP_TIMEOUT)
        except requests.exceptions.ConnectionError as ex:
            util.log_msg(
                f"FATAL ERROR: unable to connect to sht31 thermometer at url "
                f"'{self.url}'",
                mode=util.BOTH_LOG,
                func_name=1)
            raise ex
        if parameter is None:
            try:
                return response.json()
            except json.decoder.JSONDecodeError as ex:
                util.log_msg(traceback.format_exc(),
                             mode=util.BOTH_LOG,
                             func_name=1)
                if retry:
                    util.log_msg(
                        f"waiting {self.retry_delay} seconds and retrying "
                        f"SHT31 measurement one time...",
                        mode=util.BOTH_LOG,
                        func_name=1)
                    time.sleep(self.retry_delay)
                    self.get_metadata(parameter=None, retry=False)
                else:
                    raise RuntimeError("FATAL ERROR: SHT31 server "
                                       "is not responding") from ex
        else:
            try:
                return response.json()[parameter]
            except json.decoder.JSONDecodeError as ex:
                util.log_msg(traceback.format_exc(),
                             mode=util.BOTH_LOG,
                             func_name=1)
                if retry:
                    util.log_msg(
                        f"waiting {self.retry_delay} seconds and retrying "
                        f"SHT31 measurement one time...",
                        mode=util.BOTH_LOG,
                        func_name=1)
                    time.sleep(self.retry_delay)
                    self.get_metadata(parameter, retry=False)
                else:
                    raise RuntimeError("FATAL ERROR: SHT31 server "
                                       "is not responding") from ex
            except KeyError as ex:
                util.log_msg(traceback.format_exc(),
                             mode=util.BOTH_LOG,
                             func_name=1)
                if "message" in response.json():
                    util.log_msg(
                        f"WARNING in Flask response: "
                        f"'{response.json()['message']}'",
                        mode=util.BOTH_LOG,
                        func_name=1)
                if retry:
                    util.log_msg(
                        f"waiting {self.retry_delay} seconds and retrying "
                        f"SHT31 measurement one time...",
                        mode=util.BOTH_LOG,
                        func_name=1)
                    time.sleep(self.retry_delay)
                    self.get_metadata(parameter, retry=False)
                else:
                    raise KeyError(
                        f"FATAL ERROR: SHT31 server response did not contain "
                        f"key '{parameter}', raw response={response.json()}"
                        ) from ex

    def get_display_temp(self) -> float:
        """
        Return Temperature.

        inputs:
            None
        returns:
            (float): temperature in degrees.
        """
        return float(self.get_metadata(self.tempfield))

    def get_display_humidity(self) -> (float, None):
        """
        Return Humidity.

        inputs:
            None
        returns:
            (float, None): humidity in %RH, None if not supported.
        """
        raw_humidity = self.get_metadata(self.humidityfield)
        if raw_humidity is not None:
            return float(raw_humidity)
        else:
            return raw_humidity

    def get_is_humidity_supported(self) -> bool:
        """Return humidity sensor status."""
        return self.get_display_humidity() is not None

    def is_heat_mode(self) -> int:
        """Return the heat mode."""
        return 0  # not applicable

    def is_cool_mode(self) -> int:
        """Return the cool mode."""
        return 0  # not applicable

    def is_dry_mode(self) -> int:
        """Return the dry mode."""
        return 0  # not applicable

    def is_auto_mode(self) -> int:
        """Return the auto mode."""
        return 0  # not applicable

    def is_fan_mode(self) -> int:
        """Return the fan mode."""
        return 0  # not applicable

    def is_off_mode(self) -> int:
        """Return the off mode."""
        return 1  # always off

    def is_heating(self) -> int:
        """Return 1 if actively heating, else 0."""
        return 0  # not applicable

    def is_cooling(self) -> int:
        """Return 1 if actively cooling, else 0."""
        return 0  # not applicable

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return 0  # not applicable

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return 0  # not applicable

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return 0  # not applicable

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        return 1  # always on

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        return 0  # not applicable

    def is_defrosting(self):
        """Return 1 if defrosting is active, else 0."""
        return 0  # not applicable

    def is_standby(self):
        """Return 1 if standby is active, else 0."""
        return 0  # not applicable

    def get_system_switch_position(self) -> int:
        """ Return the thermostat mode.

        inputs:
            None
        returns:
            (int): thermostat mode, see tc.system_switch_position for details.
        """
        return self.system_switch_position[self.OFF_MODE]

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(
            f"display temp="
            f"{util.temp_value_with_units(self.get_display_temp())}",
            mode=util.BOTH_LOG,
            func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint()}",
                mode=util.BOTH_LOG)
            # util.log_msg("heat setpoint raw=%s" %
            #              zone.get_heat_setpoint_raw())
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint()}",
                mode=util.BOTH_LOG)
            # util.log_msg("cool setpoint raw=%s" %
            #              zone.get_cool_setpoint_raw(), mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(
            f"temporary hold until time="
            f"{self.get_temporary_hold_until_time()}",
            mode=util.BOTH_LOG)


if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=sht31_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    tc.thermostat_basic_checkout(
        sht31_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    # tc.thermostat_get_all_zone_temps(
    #     sht31_config.ALIAS,
    #     sht31_config.supported_configs["zones"],
    #     ThermostatClass,
    #     ThermostatZone)



"""
Email notifications from gmail client.

dependencies:
  environment variables must be setup:
  'GMAIL_USERNAME':  from address on gmail service
  'GMAIL_PASSWORD':  GMAIL_USERNAME password
"""
# built-in libraries
from email.mime.text import MIMEText
import smtplib
import socket
import sys
import traceback

# local libraries
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# tracing data
module_name = sys.modules[__name__]
host_name = socket.gethostname()
host_ip = socket.gethostbyname(host_name)
email_trace = (f"email sent from module '{module_name}' running on "
               f"{host_name} ({host_ip})")


def send_email_alert(to_address=None,
                     from_address=None,
                     from_password=None,
                     server_url='smtp.gmail.com',
                     server_port=465,
                     subject="", body=""):
    """
    Send an email alert on gmail.

    Email alerts are warnings so all exceptions within this module are caught
    and a status code is returned for the caller to disposition.

    inputs:
        to_address(list): list of email addresses
        from_address(str):  from gmail address
        from_password(str): password for from gmail address
        server_url(str): SMTP server URL
        server_port(int): SMTP server port number
        subject(str):  email subject text
        body(str):     email body text
    returns:
        tuple(status(int), msg(str)):  status or error code, 0 for no error
                                       and descriptive explanation.
    """
    return_status_msg_dict = {
        util.NO_ERROR: "no error",
        util.CONNECTION_ERROR: ("connection error, verify SMTP address "
                                " and port"),
        util.AUTHORIZATION_ERROR: ("authorization error, verify username "
                                   "and password"),
        util.EMAIL_SEND_ERROR: ("email send error, verify SMTP protocol "
                                "is supported by the sending and "
                                "receiving addresses"),
        util.ENVIRONMENT_ERROR: ("failed to retrieve email credentials "
                                 "from environment variable"),
    }

    # default email addresses from env variables
    if not to_address:
        buff = env.get_env_variable('GMAIL_USERNAME')
        to_address = buff["value"]
        if buff["status"] != util.NO_ERROR:
            return (buff["status"], return_status_msg_dict[buff["status"]])
    if not from_address:
        buff = env.get_env_variable('GMAIL_USERNAME')
        from_address = buff["value"]
        if buff["status"] != util.NO_ERROR:
            return (buff["status"], return_status_msg_dict[buff["status"]])
    if not from_password:
        buff = env.get_env_variable('GMAIL_PASSWORD')
        from_password = buff["value"]
        if buff["status"] != util.NO_ERROR:
            return (buff["status"], return_status_msg_dict[buff["status"]])

    status = util.NO_ERROR  # default

    # add trace
    body += f"\n\n{email_trace}"

    # build email message
    msg = MIMEText(body)
    msg['Subject'] = ["", "(unittest) "][util.unit_test_mode] + subject
    msg['From'] = from_address
    msg['To'] = to_address

    util.log_msg(f"message text={msg.as_string()}",
                 mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)

    try:
        server = smtplib.SMTP_SSL(server_url, server_port)
        util.log_msg("smtp connection successful",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
    except (ValueError,  # not sure if this exception will be raised here
            TimeoutError,  # observed on Windows for bad port
            OSError  # on AzDO with bad port
            ) as ex:
        util.log_msg(f"exception during smtp connection: {str(ex)}",
                     mode=util.BOTH_LOG, func_name=1)
        return (util.CONNECTION_ERROR, return_status_msg_dict[status])
    server.ehlo()
    try:
        server.login(from_address, from_password)
        util.log_msg(f"email account authorization for account {from_address}"
                     f" successful",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
    except (smtplib.SMTPHeloError, smtplib.SMTPAuthenticationError,
            smtplib.SMTPNotSupportedError, smtplib.SMTPException):
        util.log_msg(traceback.format_exc(),
                     mode=util.BOTH_LOG, func_name=1)
        util.log_msg(f"exception during email account authorization for "
                     f"account {from_address}",
                     mode=util.BOTH_LOG, func_name=1)
        server.close()
        return (util.AUTHORIZATION_ERROR, return_status_msg_dict[status])
    try:
        server.sendmail(from_address, to_address, msg.as_string())
        util.log_msg("mail send was successful",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
    except (smtplib.SMTPHeloError, smtplib.SMTPRecipientsRefused,
            smtplib.SMTPSenderRefused, smtplib.SMTPDataError,
            smtplib.SMTPNotSupportedError):
        util.log_msg("exception during mail send",
                     mode=util.BOTH_LOG, func_name=1)
        server.close()
        return (util.EMAIL_SEND_ERROR, return_status_msg_dict[status])
    server.close()
    util.log_msg("Email sent!",
                 mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)

    return (status, return_status_msg_dict[status])


if __name__ == "__main__":
    util.log_msg.debug = True
    send_email_alert(subject="test email alert",
                     body="this is a test of the email notification alert.")



"""
Connection to Honeywell thermoststat via TotalConnectComfort web site
using pyhtcc library.

https://pypi.org/project/pyhtcc/
"""
# built-in imports
import datetime
import os
import pprint
import time
import traceback

# local imports
from thermostatsupervisor import email_notification
from thermostatsupervisor import environment as env
from thermostatsupervisor import honeywell_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import utilities as util

# honeywell import
HONEYWELL_DEBUG = False  # debug uses local pyhtcc repo instead of pkg
if HONEYWELL_DEBUG and not env.is_azure_environment():
    mod_path = "..\\pyhtcc\\pyhtcc"
    if env.is_interactive_environment():
        mod_path = "..\\" + mod_path
    pyhtcc = env.dynamic_module_import("pyhtcc",
                                       mod_path)
else:
    import pyhtcc  # noqa E402, from path / site packages


class ThermostatClass(pyhtcc.PyHTCC, tc.ThermostatCommon):
    """Extend the PyHTCC class with additional methods."""

    def __init__(self, zone, verbose=True):
        """
        inputs:
            zone(str):  zone number.
            verbose(bool): debug flag.
        """
        # TCC server auth credentials from env vars
        self.TCC_UNAME_KEY = 'TCC_USERNAME'
        self.TCC_PASSWORD_KEY = 'TCC_PASSWORD'
        self.tcc_uname = (os.environ.get(self.TCC_UNAME_KEY, "<" +
                          self.TCC_UNAME_KEY + "_KEY_MISSING>"))
        self.tcc_pwd = (os.environ.get(
            self.TCC_PASSWORD_KEY, "<" +
            self.TCC_PASSWORD_KEY + "_KEY_MISSING>"))

        # construct the superclass
        # call both parent class __init__
        self.args = [self.tcc_uname, self.tcc_pwd]
        pyhtcc.PyHTCC.__init__(self, *self.args)
        tc.ThermostatCommon.__init__(self)

        # set tstat type and debug flag
        self.thermostat_type = honeywell_config.ALIAS
        self.verbose = verbose

        # configure zone info
        self.zone_name = int(zone)
        self.device_id = self.get_target_zone_id(self.zone_name)

    def __del__(self):
        """Clean-up session created in pyhtcc."""
        self.session.close()

    def _get_zone_device_ids(self) -> list:
        """
        Return a list of zone Device IDs.

        inputs:
            None
        returns:
            (list): all zone device ids supported.
        """
        zone_id_lst = []
        for _, zone in enumerate(self.get_zones_info()):
            zone_id_lst.append(zone['DeviceID'])
        return zone_id_lst

    def get_target_zone_id(self, zone=honeywell_config.default_zone) -> int:
        """
        Return the target zone ID.

        inputs:
            zone(int):  zone number.
        returns:
            (int): zone device id number
        """
        try:
            zone_id = self._get_zone_device_ids()[zone]
        except IndexError as ex:
            raise ValueError(f"zone '{zone}' type{type(zone)} is not a valid "
                             "choice for this Honeywell thermostat, valid "
                             "choices are: "
                             f"{self._get_zone_device_ids()}") from ex
        return zone_id

    def print_all_thermostat_metadata(self, zone, debug=False):
        """
        Return initial meta data queried from thermostat.

        inputs:
            zone(int): zone number
            debug(bool): debug flag
        returns:
            None, prints data to the stdout.
        """
        # dump all meta data
        self.get_all_metadata(zone)

        # dump uiData in a readable format
        self.exec_print_all_thermostat_metadata(
            self.get_latestdata, [zone, debug])

    def get_all_metadata(self, zone=honeywell_config.default_zone) -> dict:
        """
        Return all the current thermostat metadata.

        inputs:
          zone(int): zone number, default=honeywell_config.default_zone
        returns:
          (dict) thermostat meta data.
        """
        return_data = self.get_metadata(zone, parameter=None)
        util.log_msg(f"all meta data: {return_data}",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
        return return_data

    def get_metadata(self, zone=honeywell_config.default_zone,
                     parameter=None) -> (dict, str):
        """
        Return the current thermostat metadata settings.

        inputs:
          zone(int): zone number, default=honeywell_config.default_zone
          parameter(str): target parameter, None = all settings
        returns:
          dict if parameter=None
          str if parameter != None
        """
        zone_info_list = self.get_zones_info()
        if parameter is None:
            return_data = zone_info_list[zone]
            util.log_msg(f"zone{zone} info: {return_data}",
                         mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
            return return_data
        else:
            return_data = zone_info_list[zone].get(parameter)
            util.log_msg(f"zone{zone} parameter '{parameter}': {return_data}",
                         mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
            return return_data

    def get_latestdata(self, zone=honeywell_config.default_zone,
                       debug=False) -> dict:
        """
        Return the current thermostat latest data.

        inputs:
          zone(int): zone number, default=honeywell_config.default_zone
          debug(bool): debug flag
        returns:
          (dict) latest data from thermostat.
        """
        latest_data_dict = self.get_metadata(zone).get('latestData')
        if debug:
            util.log_msg(f"zone{zone} latestData: {latest_data_dict}",
                         mode=util.BOTH_LOG, func_name=1)
        return latest_data_dict

    def get_ui_data(self, zone=honeywell_config.default_zone) -> dict:
        """
        Return the latest thermostat ui data.

        inputs:
          zone(int): zone, default=honeywell_config.default_zone
        returns:
          (dict) ui data from thermostat.
        """
        ui_data_dict = self.get_latestdata(zone).get('uiData')
        util.log_msg(f"zone{zone} latestData: {ui_data_dict}",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
        return ui_data_dict

    def get_ui_data_param(self, zone=honeywell_config.default_zone,
                          parameter=None) -> dict:
        """
        Return the latest thermostat ui data for one specific parameter.

        inputs:
          zone(int): zone, default=honeywell_config.default_zone
          parameter(str): paramenter name
        returns:
          (dict)  # need to verify return data type.
        """
        parameter_data = self.get_ui_data(
            zone=honeywell_config.default_zone).get(parameter)
        util.log_msg(f"zone{zone} uiData parameter {parameter}: "
                     f"{parameter_data}",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG,
                     func_name=1)
        return parameter_data

    def get_zones_info(self) -> list:
        """
        Return a list of dicts corresponding with each one corresponding to a
        particular zone.

        Method overridden from base class to add exception handling.
        inputs:
            None
        returns:
            list of zone info.
        """
        return_val = []
        try:
            return_val = super().get_zones_info()
        except (pyhtcc.requests.exceptions.ConnectionError,
                pyhtcc.pyhtcc.UnexpectedError,
                pyhtcc.pyhtcc.NoZonesFoundError,
                pyhtcc.pyhtcc.UnauthorizedError,
                ) as ex:
            # force re-authenticating
            tc.connection_ok = False
            tc.connection_fail_cnt += 1
            print(traceback.format_exc())
            print(f"{util.get_function_name()}: WARNING: {ex}")

            # warning email
            time_now = (datetime.datetime.now().
                        strftime("%Y-%m-%d %H:%M:%S"))
            email_notification.send_email_alert(
                subject=(f"{self.thermostat_type} zone "
                         f"{self.zone_name}: "
                         "intermittent error "
                         "during get_zones_info()"),
                body=(f"{util.get_function_name()}: trial "
                      f"{tc.connection_fail_cnt} of "
                      f"{tc.max_connection_fail_cnt} at "
                      f"{time_now}\n{traceback.format_exc()}")
                )

            # exhausted retries, raise exception
            if tc.connection_fail_cnt > tc.max_connection_fail_cnt:
                print(f"ERRROR: exhausted {tc.max_connection_fail_cnt} retries"
                      " for get_zones_info()")
                raise ex
        except Exception as ex:
            util.log_msg(traceback.format_exc(),
                         mode=util.BOTH_LOG, func_name=1)
            print(f"ERROR: unhandled exception {ex} in get_zones_info()")
            raise ex
        else:
            # good response
            tc.connection_ok = True
            tc.connection_fail_cnt = 0  # reset

        return return_val


class ThermostatZone(pyhtcc.Zone, tc.ThermostatCommonZone):
    """Extend the Zone class with additional methods to get and set
       uiData parameters."""

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Zone constructor.

        inputs:
            Thermostat_obj(obj): Thermostat class object instance.
            verbose(bool): debug flag.
        returns:
            None
        """
        if not isinstance(Thermostat_obj.device_id, int):
            raise TypeError(
                f"device_id is type {type(Thermostat_obj.device_id)}, "
                f"expected type 'int'")

        # call both parent class __init__
        self.args = [Thermostat_obj.device_id, Thermostat_obj]
        pyhtcc.Zone.__init__(self, *self.args)
        tc.ThermostatCommonZone.__init__(self)

        # switch config for this thermostat
        self.system_switch_position[tc.ThermostatCommonZone.COOL_MODE] = 3
        self.system_switch_position[tc.ThermostatCommonZone.HEAT_MODE] = 1
        self.system_switch_position[tc.ThermostatCommonZone.OFF_MODE] = 2
        # TODO: what mode is 0 on Honeywell?

        # zone info
        self.verbose = verbose
        self.thermostat_type = honeywell_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.zone_name = Thermostat_obj.zone_name
        self.zone_name = self.get_zone_name()

        # runtime parameter defaults
        self.poll_time_sec = 10 * 60  # default to 10 minutes
        # min practical value is 2 minutes based on empirical test
        # max value was 3, higher settings will cause HTTP errors, why?
        # not showing error on Pi at 10 minutes, so changed default to 10 min.
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

    def get_zone_name(self) -> str:  # used
        """
        Refresh the cached zone information then return Name.

        inputs:
            None
        returns:
            (str): zone name
        """
        self.refresh_zone_info()
        return self.zone_info['Name']

    def get_display_temp(self) -> float:  # used
        """
        Refresh the cached zone information then return DispTemperature.

        inputs:
            None
        returns:
            (float): display temperature in degrees F.
        """
        return float(self.get_indoor_temperature_raw())

    def get_display_humidity(self) -> (float, None):
        """
        Refresh the cached zone information then return IndoorHumidity.

        inputs:
            None
        returns:
            (float, None): indoor humidity in %RH, None if not supported.
        """
        raw_humidity = self.get_indoor_humidity_raw()
        if raw_humidity is not None:
            return float(raw_humidity)
        else:
            return raw_humidity

    def get_is_humidity_supported(self) -> bool:  # used
        """
        Refresh the cached zone information and return the
          True if humidity sensor data is trustworthy.

        inputs:
            None
        returns:
            (booL): True if is in humidity sensor is available and not faulted.
        """
        self.refresh_zone_info()
        available = (self.zone_info['latestData']['uiData']
                     ['IndoorHumiditySensorAvailable'])
        not_fault = (self.zone_info['latestData']['uiData']
                     ['IndoorHumiditySensorNotFault'])
        return available and not_fault

    def is_heat_mode(self) -> int:
        """
        Refresh the cached zone information and return heat mode.

        inputs:
            None
        returns:
            (int): 1 heat mode, else 0
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.HEAT_MODE])

    def is_cool_mode(self) -> int:
        """
        Refresh the cached zone information and return the cool mode.

        inputs:
            None
        returns:
            (int): 1 if cool mode, else 0
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.COOL_MODE])

    def is_dry_mode(self) -> int:
        """
        Return the dry mode.

        inputs:
            None
        returns:
            (int): 1 if dry mode, else 0
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.DRY_MODE])

    def is_fan_mode(self) -> int:
        """
        Refresh the cached zone information and return the fan mode.

        Fan mode on Honeywell is defined as in off mode with fan set to
        on or circulate modes.
        inputs:
            None
        returns:
            (int): fan mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.OFF_MODE] and
                   (self.is_fan_on_mode() or self.is_fan_circulate_mode()))

    def is_auto_mode(self) -> int:
        """
        Return the auto mode.

        inputs:
            None
        returns:
            (int): 1 if auto mode, else 0
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.AUTO_MODE])

    def is_heating(self) -> int:
        """
        Refresh the cached zone information and return the heat active mode.
        inputs:
            None
        returns:
            (int) 1 if heating is active, else 0.
        """
        self.refresh_zone_info()
        return int(self.is_heat_mode() and
                   self.get_display_temp() < self.get_heat_setpoint_raw())

    def is_cooling(self) -> int:
        """
        Refresh the cached zone information and return the cool active mode.
        inputs:
            None
        returns:
            (int): 1 if cooling is active, else 0.
        """
        self.refresh_zone_info()
        return int(self.is_cool_mode() and
                   self.get_display_temp() > self.get_cool_setpoint_raw())

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return int(self.is_dry_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return int(self.is_auto_mode() and self.is_power_on() and
                   (self.get_cool_setpoint_raw() < self.get_display_temp() or
                    self.get_heat_setpoint_raw() > self.get_display_temp()))

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return int((self.is_fan_on() or self.is_fan_circulate_mode())
                   and self.is_power_on())

    def is_fan_circulate_mode(self):
        """Return 1 if fan is in circulate mode, else 0."""
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['fanData'][
            'fanMode'] == 2)

    def is_fan_auto_mode(self):
        """Return 1 if fan is in auto mode, else 0."""
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['fanData'][
            'fanMode'] == 0)

    def is_fan_on_mode(self):
        """Return 1 if fan is in always on mode, else 0."""
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['fanData'][
            'fanMode'] == 1)

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        self.refresh_zone_info()
        # just a guess, not sure what position 0 is yet.
        return int(self.zone_info['latestData']['uiData'][
            'SystemSwitchPosition'] > 0)

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['fanData']['fanIsRunning'])

    def get_schedule_heat_sp(self) -> int:  # used
        """
        Refresh the cached zone information and return the
        schedule heat setpoint.

        inputs:
            None
        returns:
            (int): heating set point in degrees.
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']['ScheduleHeatSp'])

    def get_schedule_cool_sp(self) -> int:
        """
        Refresh the cached zone information and return the
        schedule cool setpoint.

        inputs:
            None
        returns:
            (int): cooling set point in degrees.
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']['ScheduleCoolSp'])

    def get_is_invacation_hold_mode(self) -> bool:  # used
        """
        Refresh the cached zone information and return the
          'IsInVacationHoldMode' setting.

        inputs:
            None
        returns:
            (booL): True if is in vacation hold mode.
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']
                   ['IsInVacationHoldMode'])

    def get_vacation_hold(self) -> bool:
        """
        Refresh the cached zone information and return the
        VacationHold setting.

        inputs:
            None
        returns:
            (bool): True if vacation hold is set.
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']['VacationHold'])

    def get_vacation_hold_until_time(self) -> int:
        """
        Refresh the cached zone information and return
        the 'VacationHoldUntilTime'.
        inputs:
            None
        returns:
            (int) vacation hold time until in minutes
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']
                   ['VacationHoldUntilTime'])

    def get_temporary_hold_until_time(self) -> int:  # used
        """
        Refresh the cached zone information and return the
        'TemporaryHoldUntilTime'.

        inputs:
            None
        returns:
            (int) temporary hold time until in minutes.
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']
                   ['TemporaryHoldUntilTime'])

    def get_setpoint_change_allowed(self) -> bool:
        """
        Refresh the cached zone information and return the
        'SetpointChangeAllowed' setting.

        'SetpointChangeAllowed' will be True in heating mode,
        False in OFF mode (assume True in cooling mode too)
        inputs:
            None
        returns:
            (bool): True if set point changes are allowed.
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']
                   ['SetpointChangeAllowed'])

    def get_system_switch_position(self) -> int:  # used
        """
        Refresh the cached zone information and return the
        'SystemSwitchPosition'.

        'SystemSwitchPosition' = 1 for heat, 2 for off
        inputs:
            None
        returns:
            (int) current mode for unit, should match value
                  in self.system_switch_position
        """
        self.refresh_zone_info()
        return int(self.zone_info['latestData']['uiData']
                   ['SystemSwitchPosition'])

    def set_heat_setpoint(self, temp: int) -> None:
        """
        Set a new heat setpoint.

        This will also attempt to turn the thermostat to 'Heat'
        inputs:
            temp(int): desired temperature.
        returns:
            None
        """
        # logger.info(f"setting heat on with a target temp of: {temp}")
        return self.submit_control_changes({
            'HeatSetpoint': temp,
            'StatusHeat': 0,  # follow schedule
            'StatusCool': 0,  # follow schedule
            'SystemSwitch': self.system_switch_position[self.HEAT_MODE],
        })

    def set_cool_setpoint(self, temp: int) -> None:
        """
        Set a new cool setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature.
        returns:
            None
        """
        # logger.info(f"setting heat on with a target temp of: {temp}")
        return self.submit_control_changes({
            'CoolSetpoint': temp,
            'StatusHeat': 0,  # follow schedule
            'StatusCool': 0,  # follow schedule
            'SystemSwitch': self.system_switch_position[self.COOL_MODE],
        })

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(f"display temp="
                     f"{util.temp_value_with_units(self.get_display_temp())}",
                     mode=util.BOTH_LOG,
                     func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint()}",
                mode=util.BOTH_LOG)
            # util.log_msg("heat setpoint raw=%s" %
            #              zone.get_heat_setpoint_raw())
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint()}",
                mode=util.BOTH_LOG)
            # util.log_msg("cool setpoint raw=%s" %
            #              zone.get_cool_setpoint_raw(), mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"temporary hold until time="
                     f"{self.get_temporary_hold_until_time()}",
                     mode=util.BOTH_LOG)

    def refresh_zone_info(self, force_refresh=False) -> None:
        """
        Refresh the zone_info attribute.

        Method overridden from base class to add retry on connection errors.
        Retry up to 24 hours for extended internet outages.
        inputs:
            force_refresh(bool): not used in this method
        returns:
            None, populates self.zone_info dict.
        """
        del force_refresh  # not used
        number_of_retries = 11
        trial_number = 1
        retry_delay_sec = 60
        while trial_number < number_of_retries:
            time_now = (datetime.datetime.now().
                        strftime("%Y-%m-%d %H:%M:%S"))
            try:
                all_zones_info = self.pyhtcc.get_zones_info()
            except pyhtcc.requests.exceptions.ConnectionError:
                util.log_msg(traceback.format_exc(),
                             mode=util.BOTH_LOG,
                             func_name=1)
                msg_suffix = (
                    ["",
                     " waiting {retry_delay_sec} seconds and then " +
                     "retrying..."][trial_number < number_of_retries])
                util.log_msg(f"{time_now}: exception during refresh_zone"
                             f"_info, on trial {trial_number} of "
                             f"{number_of_retries}, probably a"
                             f" connection issue"
                             f"{msg_suffix}",
                             mode=util.BOTH_LOG, func_name=1)
                if trial_number < number_of_retries:
                    time.sleep(retry_delay_sec)
                trial_number += 1
                retry_delay_sec *= 2  # double each time.
            except Exception as ex:
                util.log_msg(traceback.format_exc(),
                             mode=util.BOTH_LOG, func_name=1)
                print(f"ERROR: unhandled exception {ex} in "
                      "refresh_zone_info()")
                raise ex
            else:
                # log the mitigated failure
                if trial_number > 1:
                    email_notification.send_email_alert(
                        subject=(f"{self.thermostat_type} zone "
                                 f"{self.zone_name}: "
                                 "intermittent connection error "
                                 "during refresh zone"),
                        body=(f"{util.get_function_name()}: trial "
                              f"{trial_number} of {number_of_retries} at "
                              f"{time_now}")
                        )
                for zone_data in all_zones_info:
                    if zone_data['DeviceID'] == self.device_id:
                        pyhtcc.logger.debug(f"Refreshed zone info for \
                                           {self.device_id}")
                        self.zone_info = zone_data
                        self.last_fetch_time = time.time()
                        return

        # log fatal failure
        email_notification.send_email_alert(
            subject=(f"{self.thermostat_type} zone {self.zone_name}: "
                     "fatal connection error during refresh zone"),
            body=(f"{util.get_function_name()}: trial {trial_number} of "
                  f"{number_of_retries} at {time_now}"))
        # raise pyhtcc.ZoneNotFoundError(f"Missing device: {self.device_id}")

        # trigger a forced reconnection
        tc.connection_ok = False


# add default requests session default timeout to prevent TimeoutExceptions
# see ticket #93 for details
# pylint: disable=wrong-import-order,wrong-import-position
from requests.adapters import HTTPAdapter  # noqa E402

# network timeout limit
# 6s upper is 1.9 on pi4 and laptop
# 6s upper is 2.17 on Azure pipeline
HTTP_TIMEOUT = 2.5  # 6 sigma limit in seconds


class TimeoutHTTPAdapter(HTTPAdapter):
    """Override TimeoutHTTPAdapter to include timeout parameter.
    """

    def __init__(self, *args, **kwargs):
        self.timeout = HTTP_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):  # pylint: disable=arguments-differ
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=honeywell_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    _, Zone = tc.thermostat_basic_checkout(
        honeywell_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    tc.thermostat_get_all_zone_temps(
        honeywell_config.ALIAS,
        honeywell_config.supported_configs["zones"],
        ThermostatClass,
        ThermostatZone)

    # measure thermostat response time
    MEASUREMENTS = 30
    print(f"Thermostat response times for {MEASUREMENTS} measurements...")
    meas_data = Zone.measure_thermostat_response_time(
        MEASUREMENTS, func=Zone.pyhtcc.get_zones_info)
    ppp = pprint.PrettyPrinter(indent=4)
    ppp.pprint(meas_data)



"""
Sample setup.py file
"""

# built-in libraries
import codecs
import os

# third party libraries
from setuptools import setup

# local imports

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()


def read(rel_path):
    """
    Open and read file.

    inputs:
        rel_path(str): path to file containing version str.
    returns:
        (file pointer): pointer to open file.
    """
    print(f"base path={here}")
    print(f"relative path={rel_path}")
    dir_list = os.listdir(here)
    print(f"files in base path: {dir_list}")
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    """
    Get package version from specified file.

    inputs:
        rel_path(str): path to file containing version str.
    returns:
        (str): version string.
    """
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    version=get_version("thermostatsupervisor/__init__.py"),
)



"""
Unit test module for thermostat_api.py.
"""
# built-in imports
import os
import sys
import unittest

# local imports
from thermostatsupervisor import emulator_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc

thermostat_type = emulator_config.ALIAS
zone_name = emulator_config.default_zone_name


class Test(utc.UnitTest):
    """Test functions in thermostat_api.py."""
    tstat = thermostat_type
    zone_name = zone_name

    def setUp(self):
        super().setUp()
        self.setup_mock_thermostat_zone()
        api.thermostats[self.thermostat_type] = {  # dummy unit test thermostat
            "required_env_variables": {
                "GMAIL_USERNAME": None,
                "GMAIL_PASSWORD": None,
            },
        }

    def tearDown(self):
        self.teardown_mock_thermostat_zone()
        super().tearDown()

    def test_verify_required_env_variables(self):
        """
        Verify verify_required_env_variables() passes in nominal
        condition and fails with missing key.
        """
        missing_key = "agrfg_"  # bogus key should be missing

        # nominal condition, should pass
        print("testing nominal condition, will pass if gmail keys are present")
        self.assertTrue(api.verify_required_env_variables(
            self.thermostat_type, "0"),
            "test failed because one or more gmail keys "
            "are missing")

        # missing key, should raise exception
        print("testing for with missing key 'unit_test', should fail")
        api.thermostats[self.thermostat_type][
            "required_env_variables"][missing_key] = "bogus_value"
        try:
            self.assertFalse(api.verify_required_env_variables(
                self.thermostat_type, "0"),
                f"test passed with missing key '{missing_key}', "
                f"should have failed")
        except KeyError:
            print("KeyError raised as expected for missing key")
        else:
            self.fail("expected KeyError but exception did not occur")
        finally:
            api.thermostats[self.thermostat_type][
                "required_env_variables"].pop(missing_key)

    def test_load_hardware_library(self):
        """
        Verify load_hardware_library() runs without error
        """
        # test successful case
        pkg = api.load_hardware_library(emulator_config.ALIAS)
        print(f"default thermostat returned package type {type(pkg)}")
        self.assertTrue(isinstance(pkg, object),
                        f"dynamic_module_import() returned type({type(pkg)}), "
                        f"expected an object")
        print(f"package name={pkg.__name__}")
        del sys.modules[pkg.__name__]
        del pkg

        # test failing case
        with self.assertRaises(KeyError):
            print("attempting to access 'bogus' dictionary key, "
                  "expect exception...")
            pkg = api.load_hardware_library("bogus")
            print(f"'bogus' returned package type {type(pkg)}, exception "
                  f"should have been raised")
            del pkg
        print("test passed")

    def test_max_measurement_count_exceeded(self):
        """
        Verify max_measurement_count_exceeded() runs as expected.
        """
        test_cases = {
            "within_range": {"measurement": 13, "max_measurements": 14,
                             "exp_result": False},
            "at_range": {"measurement": 17, "max_measurements": 17,
                         "exp_result": False},
            "over_range": {"measurement": 15, "max_measurements": 14,
                           "exp_result": True},
            "default": {"measurement": 13, "max_measurements": None,
                        "exp_result": False},
        }
        # backup max_measurements
        api.uip = api.UserInputs(self.unit_test_argv, "unit test parser",
                                 self.tstat,
                                 self.zone_name)
        max_measurement_bkup = api.uip.get_user_inputs(
            api.uip.zone_name, api.input_flds.measurements)
        try:
            for test_case, parameters in test_cases.items():
                api.uip.set_user_inputs(api.uip.zone_name,
                                        api.input_flds.measurements,
                                        parameters["max_measurements"])
                act_result = api.uip.max_measurement_count_exceeded(
                    parameters["measurement"])
                exp_result = parameters["exp_result"]
                self.assertEqual(exp_result, act_result,
                                 f"test case '{test_case}', "
                                 f"expected={exp_result}, actual={act_result}")
        finally:
            # restore max masurements
            api.uip.set_user_inputs(api.uip.zone_name,
                                    api.input_flds.measurements,
                                    max_measurement_bkup)


class RuntimeParameterTest(utc.RuntimeParameterTest):
    """API Runtime parameter tests."""

    mod = api  # module to test

    script = os.path.realpath(__file__)
    thermostat_type = thermostat_type
    # parent_key = zone_name  # aka zone_name in this context
    zone = 0
    poll_time_sec = 9
    connection_time_sec = 90
    tolerance = 3
    target_mode = "HEAT_MODE"
    measurements = 1
    input_file = utc.unit_test_argv_file

    # fields for testing, mapped to class variables.
    # (value, field name)
    test_fields = [
        (script, os.path.realpath(__file__)),
        (thermostat_type, api.input_flds.thermostat_type),
        (zone, api.input_flds.zone),
        (poll_time_sec, api.input_flds.poll_time),
        (connection_time_sec, api.input_flds.connection_time),
        (tolerance, api.input_flds.tolerance),
        (target_mode, api.input_flds.target_mode),
        (measurements, api.input_flds.measurements),
    ]
    # test case with input file
    test_fields_with_file = [
        (script, os.path.realpath(__file__)),
        (thermostat_type, api.input_flds.thermostat_type),
        (zone, api.input_flds.zone),
        (poll_time_sec, api.input_flds.poll_time),
        (connection_time_sec, api.input_flds.connection_time),
        (tolerance, api.input_flds.tolerance),
        (target_mode, api.input_flds.target_mode),
        (measurements, api.input_flds.measurements),
        (input_file, api.input_flds.input_file),
        ]


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Common functions used in multiple unit tests.
"""
# global imports
import argparse
import os
import pprint
import sys
import unittest
from unittest.mock import patch

# third party imports
from str2bool import str2bool

# local imports
from thermostatsupervisor import emulator_config
from thermostatsupervisor import honeywell_config
from thermostatsupervisor import supervise as sup
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# enable modes
ENABLE_FUNCTIONAL_INTEGRATION_TESTS = True  # enable func int tests
ENABLE_PERFORMANCE_INTEGRATION_TESTS = False and \
    not env.is_azure_environment()  # enable performance int tests
ENABLE_SUPERVISE_INTEGRATION_TESTS = True  # enable supervise int tests
ENABLE_FLASK_INTEGRATION_TESTS = True  # enable flask int tests
ENABLE_KUMOLOCAL_TESTS = False  # Kumolocal is local net only
ENABLE_MMM_TESTS = False  # mmm50 is local net only
ENABLE_SHT31_TESTS = False  # sht31 can fail on occasion
ENABLE_BLINK_TESTS = True and \
    not env.is_azure_environment()  # Blink cameras, TODO #638


# generic argv list for unit testing
unit_test_emulator = emulator_config.argv

unit_test_sht31 = ["supervise.py",  # module
                   "sht31",  # thermostat
                   ["99", "1"][env.is_azure_environment()],  # zone
                   "19",  # poll time in sec
                   "359",  # reconnect time in sec
                   "3",  # tolerance
                   "OFF_MODE",  # thermostat mode
                   "2",  # number of measurements
                   ]

unit_test_honeywell = honeywell_config.argv

unit_test_argv = unit_test_emulator
unit_test_argv_file = ".//tests//unit_test_argv.txt"


class PatchMeta(type):
    """A metaclass to patch all inherited classes."""

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # mock argv to prevent azure runtime args from polluting test.
        patch.object(*cls.patch_args)(cls)  # noqa e501, pylint:disable=undefined-variable


class UnitTest(unittest.TestCase, metaclass=PatchMeta):
    """Extensions to unit test framework."""

    # mock argv to prevent unit test runtime arguments from polluting tests
    # that use argv parameters.  Azure pipelines is susceptible to this issue.
    __metaclass__ = PatchMeta
    patch_args = (sys, 'argv', [os.path.realpath(__file__)])

    mod = None  # imported module
    thermostat_type = None  # thermostat type
    zone = None  # zone number
    Thermostat = None  # thermostat instance
    Zone = None  # zone instance

    user_inputs_backup = None
    is_off_mode_bckup = None

    def setUp(self):
        """Default setup method."""
        self.print_test_name()
        self.unit_test_argv = unit_test_argv
        self.thermostat_type = unit_test_argv[1]
        self.zone = unit_test_argv[2]
        util.unit_test_mode = True

    def tearDown(self):
        """Default teardown method."""
        self.print_test_result()

    def setup_thermostat_zone(self):
        """
        Create a Thermostat and Zone instance for unit testing if needed.

        This function is called at the beginning of integration tests.
        """
        # parse runtime arguments
        api.uip = api.UserInputs(self.unit_test_argv)

        # create new Thermostat and Zone instances
        if self.Thermostat is None and self.Zone is None:
            util.log_msg.debug = True  # debug mode set
            thermostat_type = api.uip.get_user_inputs(
                api.uip.zone_name,
                api.input_flds.thermostat_type)
            zone = api.uip.get_user_inputs(api.uip.zone_name,
                                           api.input_flds.zone)

            # create class instances
            self.Thermostat, self.Zone = tc.create_thermostat_instance(
                thermostat_type, zone, self.mod.ThermostatClass,
                self.mod.ThermostatZone)

        # return instances
        return self.Thermostat, self.Zone

    def setup_mock_thermostat_zone(self):
        """Setup mock thermostat settings."""
        api.thermostats[self.thermostat_type] = {  # dummy unit test thermostat
            "required_env_variables": {
                "GMAIL_USERNAME": None,
                "GMAIL_PASSWORD": None,
            },
        }
        self.unit_test_argv = unit_test_argv  # use defaults
        self.user_inputs_backup = getattr(api.uip, "user_inputs", None)
        # parse runtime arguments
        api.uip = api.UserInputs(self.unit_test_argv)

        self.Thermostat = tc.ThermostatCommon()
        self.Zone = tc.ThermostatCommonZone()
        self.Zone.update_runtime_parameters()
        self.Zone.current_mode = self.Zone.OFF_MODE
        self.is_off_mode_bckup = self.Zone.is_off_mode
        self.Zone.is_off_mode = lambda *_, **__: 1

    def teardown_mock_thermostat_zone(self):
        """Tear down the mock thermostat settings."""
        del api.thermostats[self.thermostat_type]
        api.uip.user_inputs = self.user_inputs_backup
        self.Zone.is_off_mode = self.is_off_mode_bckup

    def print_test_result(self):
        """Print unit test result to console."""
        if hasattr(self._outcome, 'errors'):  # Python 3.4 - 3.10
            # These two methods have no side effects
            result = self.defaultTestResult()
            self._feedErrorsToResult(result, self._outcome.errors)
        elif hasattr(self._outcome, 'result'):  # python 3.11
            # These two methods have no side effects
            result = self._outcome.result
        else:  # Python 3.2 - 3.3 or 3.0 - 3.1 and 2.7
            raise OSError(
                "this code is designed to work on Python 3.4+")
            # result = getattr(self, '_outcomeForDoCleanups',
            #                 self._resultForDoCleanups)
        error = self.list2reason(result.errors)
        failure = self.list2reason(result.failures)
        result_ok = not error and not failure

        # Demo:   report short info immediately (not important)
        if not result_ok:
            typ, text = ('ERROR', error) if error else ('FAIL', failure)
            msg = [x for x in text.split('\n')[1:] if not x.startswith(' ')][0]
            print(f"\n{typ}: {self.id()}\n     {msg}")

    def list2reason(self, exc_list):
        """Parse reason from list."""
        if exc_list and exc_list[-1][0] is self:
            return exc_list[-1][1]
        else:
            return None

    def print_test_name(self):
        """Print out the unit test name to the console."""
        print("\n")
        print("-" * 60)
        print(f"testing '{self.id()}'")  # util.get_function_name(2))
        print("-" * 60)


class IntegrationTest(UnitTest):
    """Common integration test framework."""

    Thermostat = None  # Thermostat object instance
    Zone = None  # Zone object instance
    mod = None  # module object
    mod_config = None  # config object
    unit_test_argv = []  # populated during setup
    timeout_limit = None
    timing_measurements = None
    timing_func = None
    temp_stdev_limit = None
    temp_repeatability_measurements = None
    humidity_stdev_limit = None
    humidity_repeatability_measurements = None
    poll_interval_sec = None

    def setup_common(self):
        """Test attributes common to all integration tests."""
        self.timeout_limit = 30  # 6 sigma timing upper limit in sec.
        self.timing_measurements = 10  # number of timing measurements.
        self.timing_func = None  # function used for timing measurement.
        self.temp_stdev_limit = 1  # 1 sigma temp repeatability limit in F
        self.temp_repeatability_measurements = 10  # number of temp msmts.
        self.humidity_stdev_limit = 1  # 1 sigma humid repeatability limit %RH
        self.humidity_repeatability_measurements = 10  # number of humid msmts.
        self.poll_interval_sec = 0  # delay between repeat measurements

    def setup_thermostat_zone(self):
        """
        Create a Thermostat and Zone instance for integration testing.

        This function is called at the beginning of integration tests.
        If an existing thermostat/zone exists from previous test this function
        will use the existing instances.
        """
        # parse runtime arguments
        api.uip = api.UserInputs(self.unit_test_argv)

        # create new Thermostat and Zone instances
        if self.Thermostat is None and self.Zone is None:
            util.log_msg.debug = True  # debug mode set
            thermostat_type = api.uip.get_user_inputs(
                api.uip.zone_name,
                api.input_flds.thermostat_type)
            zone = api.uip.get_user_inputs(api.uip.zone_name,
                                           api.input_flds.zone)

            # create class instances
            self.Thermostat, self.Zone = tc.create_thermostat_instance(
                thermostat_type, zone, self.mod.ThermostatClass,
                self.mod.ThermostatZone)

        # update runtime parameters
        self.Zone.update_runtime_parameters()

        # return instances
        return self.Thermostat, self.Zone


@unittest.skipIf(not ENABLE_FUNCTIONAL_INTEGRATION_TESTS,
                 "functional integration tests are disabled")
class FunctionalIntegrationTest(IntegrationTest):
    """Functional integration tests."""

    metadata_field = None  # thermostat-specific
    metadata_type = str  # thermostat-specific

    def test_a_thermostat_basic_checkout(self):
        """
        Verify thermostat_basic_checkout on target thermostat.

        This test also creates the class instances so it should be run
        first in the integration test sequence.
        """
        api.uip = api.UserInputs(self.unit_test_argv)

        IntegrationTest.Thermostat, IntegrationTest.Zone = \
            tc.thermostat_basic_checkout(
                api.uip.get_user_inputs(api.uip.zone_name,
                                        api.input_flds.thermostat_type),
                api.uip.get_user_inputs(api.uip.zone_name,
                                        api.input_flds.zone),
                self.mod.ThermostatClass, self.mod.ThermostatZone
            )

    def test_report_heating_parameters(self):
        """
        Verify report_heating_parameters().
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        for test_case in self.mod_config.supported_configs["modes"]:
            print("-" * 80)
            print(f"test_case='{test_case}'")
            with patch.object(self.Zone, "get_system_switch_position",  # noqa e501, pylint:disable=undefined-variable
                              return_value=self.Zone.system_switch_position[
                                  test_case]):
                self.Zone.report_heating_parameters(
                    self.Zone.system_switch_position[test_case])
            print("-" * 80)

    def test_get_all_meta_data(self):
        """
        Verify get_all_metadata().
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        expected_return_type = dict
        metadata = self.Thermostat.get_all_metadata(
            zone=self.Thermostat.zone_name)
        self.assertTrue(
            isinstance(
                metadata,
                expected_return_type),
            f"metadata is type '{type(metadata)}', "
            f"expected type '{expected_return_type}'")

    def test_get_meta_data(self):
        """
        Verify get_metadata().
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # test None case
        parameter = None
        expected_return_type = dict
        metadata = self.Thermostat.get_metadata(
            zone=self.Thermostat.zone_name,
            parameter=parameter)
        self.assertTrue(
            isinstance(
                metadata,
                expected_return_type),
            f"parameter='{parameter}', metadata is type '{type(metadata)}', "
            f"expected type '{expected_return_type}'")

        # test parameter case
        parameter = self.metadata_field
        expected_return_type = self.metadata_type
        metadata = self.Thermostat.get_metadata(
            zone=self.Thermostat.zone_name,
            parameter=parameter)
        self.assertTrue(
            isinstance(
                metadata,
                expected_return_type),
            f"parameter='{parameter}', value={metadata}, metadata is type "
            f"'{type(metadata)}', expected type '{expected_return_type}'")


@unittest.skipIf(not ENABLE_SUPERVISE_INTEGRATION_TESTS,
                 "supervise integration test is disabled")
class SuperviseIntegrationTest(IntegrationTest):
    """Supervise integration tests common to all thermostat types."""

    def test_supervise(self):
        """
        Verify supervisor loop on target thermostat.
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        return_status = sup.exec_supervise(
            debug=True, argv_list=self.unit_test_argv)
        self.assertTrue(return_status,
                        f"return status={return_status}, expected True")


@unittest.skipIf(not ENABLE_PERFORMANCE_INTEGRATION_TESTS,
                 "performance integration tests are disabled")
class PerformanceIntegrationTest(IntegrationTest):
    """Performance integration tests common to all thermostat types."""

    def test_network_timing(self):
        """
        Verify network timing..
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # measure thermostat response time
        measurements = self.timing_measurements
        print(f"{self.Zone.thermostat_type} Thermostat zone "
              f"{self.Zone.zone_name} response times for {measurements} "
              f"measurements...")
        meas_data = \
            self.Zone.measure_thermostat_response_time(measurements)
        print("network timing stats (sec)")
        ppp = pprint.PrettyPrinter(indent=4)
        ppp.pprint(meas_data)

        # fail test if any measurement fails the limit.
        self.assertTrue(
            meas_data['max'] <= self.timeout_limit,
            f"max value observed ({meas_data['max']}) is greater than timout"
            f" setting ({self.timeout_limit})")

        # fail test if thermostat timing margin is poor vs. 6 sigma value
        self.assertTrue(
            meas_data['6sigma_upper'] <= self.timeout_limit,
            f"6 sigma timing margin ({meas_data['6sigma_upper']}) is greater "
            f"than timout setting ({self.timeout_limit})")

    def test_temperature_repeatability(self):
        """
        Verify temperature repeatability.
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # measure thermostat temp repeatability
        measurements = self.temp_repeatability_measurements
        print(f"{self.Zone.thermostat_type} Thermostat zone "
              f"{self.Zone.zone_name} temperature repeatability for "
              f"{measurements} measurements with {self.poll_interval_sec} "
              f"sec delay between each measurement...")
        meas_data = self.Zone.measure_thermostat_repeatability(
            measurements, self.poll_interval_sec)
        print("temperature repeatability stats (deg F)")
        ppp = pprint.PrettyPrinter(indent=4)
        ppp.pprint(meas_data)

        # fail test if thermostat temp repeatability is poor
        act_val = meas_data['stdev']
        self.assertTrue(
            act_val <= self.temp_stdev_limit,
            f"temperature stdev ({act_val}) is greater than temp repeatability"
            f" limit ({self.temp_stdev_limit})")

    def test_humidity_repeatability(self):
        """
        Verify humidity repeatability.
        """
        # setup class instances
        self.Thermostat, self.Zone = self.setup_thermostat_zone()

        # check for humidity support
        if not self.Zone.get_is_humidity_supported():
            print("humidity not supported on this thermostat, exiting")
            return

        # measure thermostat humidity repeatability
        measurements = self.temp_repeatability_measurements
        print(f"{self.Zone.thermostat_type} Thermostat zone "
              f"{self.Zone.zone_name} humidity repeatability for "
              f"{measurements} measurements with {self.poll_interval_sec} "
              f"sec delay betweeen each measurement...")
        meas_data = self.Zone.measure_thermostat_repeatability(
            measurements, self.poll_interval_sec,
            self.Zone.get_display_humidity)
        print("humidity repeatability stats (%RH)")
        ppp = pprint.PrettyPrinter(indent=4)
        ppp.pprint(meas_data)

        # fail test if thermostat humidity repeatability is poor
        act_val = meas_data['stdev']
        self.assertTrue(
            act_val <= self.humidity_stdev_limit,
            f"humidity stdev ({act_val}) is greater than humidity "
            f"repeatability limit ({self.humidity_stdev_limit})")


class RuntimeParameterTest(UnitTest):
    """Runtime parameter tests."""

    uip = None
    mod = None
    test_fields = []  # placeholder, will be populated by child classes
    test_fields_with_file = None  # placeholder, will be populated by child
    parent_key = util.default_parent_key  # will be updated during inheritance.

    def get_test_list(self, test_fields):
        """
        Return the test list with string elements.

        inputs:
            test_fields(list of tuples): test field mapping.
        returns:
            (list): list of test fields."""
        test_list = []
        for k, _ in test_fields:
            test_list.append(k)
        return test_list

    def get_expected_vals_dict(self, parent_key):
        """
        Return the expected values dictionary.

        inputs:
            parent_key(str, int): parent key for dict.
        """
        expected_values = {}
        expected_values[parent_key] = {}
        # element 0 (script) is omitted from expected_values dict.
        for x in range(1, len(self.test_fields)):
            expected_values[parent_key][
                self.test_fields[x][1]] = self.test_fields[x][0]
        return expected_values

    def get_named_list(self, test_fields, flag):
        """
        Return the named parameter list.

        inputs:
            test_fields(list of tuples): test field mapping.
            flag(str): flag.
        returns:
            (list): named parameter list
        """
        test_list_named_flag = []
        # script placeholder for 0 element
        test_list_named_flag.append(test_fields[0][1])

        # element 0 (script) is omitted from expected_values dict.
        for field in range(1, len(test_fields)):
            test_list_named_flag.append(self.uip.get_user_inputs(
                list(self.uip.user_inputs.keys())[0],
                test_fields[field][1], flag) + "=" +
                str(test_fields[field][0]))
        return test_list_named_flag

    def parse_user_inputs_dict(self, key):
        """
        Parse the user_inputs_dict into list matching
        order of test_list.

        inputs:
            key(str): key within user_inputs to parse.
        returns:
            (list) of actual values.
        """
        actual_values = []
        for x in range(0, len(self.test_fields)):
            actual_values.append(self.uip.get_user_inputs(
                list(self.uip.user_inputs.keys())[0],
                key, self.test_fields[x][1]))
        return actual_values

    def setUp(self):
        self.print_test_name()
        util.log_msg.file_name = "unit_test.txt"
        self.initialize_user_inputs()

    def tearDown(self):
        self.print_test_result()

    def verify_parsed_values(self, parent_key=None):
        """
        Verify values were parsed correctly by comparing to expected values.

        inputs:
            parent_key(str, int): parent_key for dict.
        """
        if self.uip.using_input_file:
            expected_values = self.uip.user_inputs_file
        else:
            expected_values = self.get_expected_vals_dict(parent_key)

        for parent_key, child_dict in expected_values.items():
            for c_key, _ in child_dict.items():
                self.assertEqual(
                    expected_values[parent_key][c_key],
                    self.uip.get_user_inputs(parent_key, c_key),
                    f"expected({type(expected_values[parent_key][c_key])})"
                    f" {expected_values[parent_key][c_key]} != "
                    f"actual("
                    f"{type(self.uip.get_user_inputs(parent_key, c_key))})"
                    f" {self.uip.get_user_inputs(parent_key, c_key)}")

    def initialize_user_inputs(self):
        """
        Re-initialize user_inputs dict.
        """
        print(f"{util.get_function_name()}:initializing user_inputs with "
              "defaults...")
        self.uip = self.mod.UserInputs(suppress_warnings=True)
        print(f"{util.get_function_name()}:user_inputs have been initialized.")
        self.uip.suppress_warnings = False  # reset back to default
        for parent_key in self.uip.user_inputs:
            for child_key in self.uip.user_inputs[parent_key]:
                self.uip.set_user_inputs(parent_key, child_key, None)

    def test_parse_argv_list(self):
        """
        Verify test parse_argv_list() returns expected
        values when input known values.
        """
        test_list = self.get_test_list(self.test_fields)
        print(f"test_list={test_list}")
        self.uip = self.mod.UserInputs(test_list, "unit test parser")
        print(f"user_inputs={self.uip.user_inputs}")
        self.verify_parsed_values(self.uip.default_parent_key)

    def test_parser(self):
        """
        Generic test for argparser.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('-a', type=int)
        # argv = '-a 1'.split()  # or ['-a','1','foo']
        # argv = ["-a 1", "--b 2"]  # double dash doesn't work yet.
        argv = ["-a=1"]
        args = parser.parse_args(argv)
        assert (args.a == 1)
        # assert(args.b == 2)

    def test_parser_input_file(self):
        """
        Generic test for argparser with input file.
        """
        input_file = unit_test_argv_file
        parser = argparse.ArgumentParser()
        parser.add_argument('-f',
                            type=argparse.FileType('r',
                                                   encoding='UTF-8')
                            )
        argv = ["-f" + input_file]  # space after sflag is appended onto str
        args = parser.parse_args(argv)
        print("args returned: %s" %
              (' '.join(f'{k}={v}' for k, v in vars(args).items())))
        # assert(args.thermostat_type == "emulator")

    def is_valid_file(self, parser, arg):
        """
        Verify file input is valid.

        inputs:
            arg(str): file name with path.
        returns:
            open file handle
        """
        arg = arg.strip()  # remove any leading spaces
        if not os.path.exists(arg):
            parser.error(f"The file {os.path.abspath(arg)} does not exist!")
        else:
            return open(arg, 'rt', encoding="utf8")  # return a file handle

    def test_parse_named_arguments_sflag(self):
        """
        Verify test parse_named_arguments() returns expected
        values when input known values with sflag.
        """
        # build the named sflag list
        self.uip = self.mod.UserInputs()
        named_sflag_list = self.get_named_list(self.test_fields, "sflag")
        print(f"sflag_list={named_sflag_list}")

        # clear out user inputs
        self.initialize_user_inputs()

        # parse named sflag list
        # override parent key since list input is provided.
        print("parsing named argument list")
        self.uip = self.mod.UserInputs(
            named_sflag_list, "unittest parsing named sflag arguments")
        print(f"in test default parent key={self.uip.default_parent_key}")
        self.verify_parsed_values(self.uip.default_parent_key)

    def test_parse_named_arguments_lflag(self):
        """
        Verify test parse_named_arguments() returns expected
        values when input known values with sflag.
        """
        return  # not yet working

        # build the named sflag list
        self.uip = self.mod.UserInputs()
        named_lflag_list = self.get_named_list(self.test_fields, "lflag")
        print(f"lflag_list={named_lflag_list}")

        # clear out user inputs
        self.initialize_user_inputs()

        # parse named sflag list
        # override parent key since list input is provided.
        self.uip = self.mod.UserInputs(
            named_lflag_list, "unittest parsing named sflag arguments")
        self.verify_parsed_values(util.default_parent_key)

    def parse_named_arguments(self, test_list, label_str):
        """
        Verify test parse_named_arguments() returns expected
        values when input known values.

        inputs:
            test_list(list): list of named arguments
            label_str(str): description pass-thru
        """
        print(f"testing named arg list='{test_list}")
        self.uip = self.mod.UserInputs(test_list, label_str)
        print(f"user_inputs={self.uip.user_inputs}")
        self.verify_parsed_values(util.default_parent_key)

    def test_parse_runtime_parameters(self):
        """
        Test the upper level function for parsing.
        """
        self.parse_runtime_parameters(self.test_fields)

    def test_parse_runtime_parameters_from_file(self):
        """
        Test the upper level function for parsing.
        """
        if self.test_fields_with_file is not None:
            self.parse_runtime_parameters(self.test_fields_with_file)
        else:
            print("self.test_list_file has not been setup yet, "
                  "skipping this test.")

    def parse_runtime_parameters(self, test_fields):
        """
        Test the upper level function for parsing.

        inputs:
            test_fields(list of tuples): test field mapping.
        """
        print("test 1, user_inputs=None, will raise error")
        self.uip = self.mod.UserInputs()
        try:
            self.uip.user_inputs = None
            with self.assertRaises(ValueError):
                self.uip.parse_runtime_parameters(argv_list=None)
        finally:
            self.uip = self.mod.UserInputs()
        print("passed test 1")

        # initialize parser so that lower level functions can be tested.
        self.uip = self.mod.UserInputs(help_description="unit test parsing")

        print("test 2, input list, will parse list")
        self.initialize_user_inputs()
        test_list = self.get_test_list(test_fields)
        print(f"test2 test_list={test_list}")
        self.uip.parse_runtime_parameters(
            argv_list=test_list)
        self.verify_parsed_values(self.uip.default_parent_key)

        print("test 3, input named parameter list, will parse list")
        self.initialize_user_inputs()
        self.uip.parse_runtime_parameters(
            argv_list=self.get_named_list(test_fields, "sflag"))
        self.verify_parsed_values(self.uip.default_parent_key)

        print("test 4, input dict, will parse sys.argv argument list")
        self.initialize_user_inputs()
        with patch.object(sys, 'argv', self.get_test_list(test_fields)):  # noqa e501, pylint:disable=undefined-variable
            self.uip.parse_runtime_parameters(argv_list=None)
        self.verify_parsed_values(self.uip.default_parent_key)

        print("test 5, input dict, will parse sys.argv named args")
        self.initialize_user_inputs()
        with patch.object(sys, 'argv', self.get_named_list(test_fields, "sflag")):  # noqa e501, pylint:disable=undefined-variable
            self.uip.parse_runtime_parameters()
        self.verify_parsed_values(self.uip.default_parent_key)

    def test_validate_argv_inputs(self):
        """
        Verify validate_argv_inputs() works as expected.
        """
        test_cases = {
            "fail_missing_value": {
                "value": None,
                "type": int,
                "default": 1,
                "valid_range": range(0, 4),
                "expected_value": 1,
                "required": False,
                },
            "fail_datatype_error": {
                "value": "5",
                "type": int,
                "default": 2,
                "valid_range": range(0, 10),
                "expected_value": 2,
                "required": False,
                },
            "fail_out_of_range_int": {
                "value": 6,
                "type": int,
                "default": 3,
                "valid_range": range(0, 3),
                "expected_value": 3,
                "required": False,
                },
            "fail_out_of_range_str": {
                "value": "6",
                "type": str,
                "default": "4",
                "valid_range": ["a", "b"],
                "expected_value": "4",
                "required": False,
                },
            "in_range_int": {
                "value": 7,
                "type": int,
                "default": 4,
                "valid_range": range(0, 10),
                "expected_value": 7,
                "required": False,
                },
            "in_range_str": {
                "value": "8",
                "type": str,
                "default": "5",
                "valid_range": ["a", "8", "abc"],
                "expected_value": "8",
                "required": False,
                },
            }

        child_key = "test_case"
        for test_case, test_dict in test_cases.items():
            print(f"test case='{test_case}'")
            result_dict = self.uip.validate_argv_inputs({self.parent_key: {
                child_key: test_dict}})
            actual_value = result_dict[self.parent_key][child_key]["value"]

            if "fail_" in test_case:
                expected_value = result_dict[self.parent_key][child_key][
                    "default"]
            else:
                expected_value = (result_dict[self.parent_key][child_key]
                                  ["expected_value"])

            self.assertEqual(expected_value, actual_value,
                             f"test case ({test_case}), "
                             f"expected={expected_value}, "
                             f"actual={actual_value}")


# user input fields
BOOL_FLD = "bool_field"
INT_FLD = "int_field"
FLOAT_FLD = "float_field"
STR_FLD = "str_field"
REQUIRED_FLD = "required_field"
INPUT_FILE_FLD = "input_file"
uip = {}


class UserInputs(util.UserInputs):
    """Manage runtime arguments for generic unit testing."""

    def __init__(self, argv_list=None, help_description=None,
                 suppress_warnings=False):
        """
        UserInputs constructor for generic unit testing.

        inputs:
            argv_list(list): override runtime values
            help_description(str): description field for help text
            suppress_warnings(bool): suppress warnings
        """
        self.argv_list = argv_list
        self.help_description = help_description
        self.suppress_warnings = suppress_warnings
        self.user_inputs_parent = util.default_parent_key

        # initialize parent class
        super().__init__(argv_list, help_description, suppress_warnings)

    def initialize_user_inputs(self, parent_keys=None):
        """
        Populate user_inputs dict.
        """
        # define the user_inputs dict.
        self.user_inputs = {self.user_inputs_parent: {
            BOOL_FLD: {
                "order": 1,  # index in the argv list
                "value": None,
                "type": lambda x: bool(str2bool(
                    str(x).strip())),
                "default": False,
                "valid_range": [True, False, 1, 0],
                "sflag": "-b",
                "lflag": "--" + BOOL_FLD,
                "help": "bool input parameter",
                "required": False,
                },
            INT_FLD: {
                "order": 2,  # index in the argv list
                "value": None,
                "type": int,
                "default": 49,
                "valid_range": range(0, 99),
                "sflag": "-i",
                "lflag": "--" + INT_FLD,
                "help": "int input parameter",
                "required": False,
                },
            FLOAT_FLD: {
                "order": 3,  # index in the argv list
                "value": None,
                "type": float,
                "default": 59.0,
                "valid_range": None,
                "sflag": "-l",
                "lflag": "--" + FLOAT_FLD,
                "help": "float input parameter",
                "required": False,
                },
            STR_FLD: {
                "order": 4,  # index in the argv list
                "value": None,
                "type": str,
                "default": "this is a string",
                "valid_range": None,
                "sflag": "-s",
                "lflag": "--" + STR_FLD,
                "help": "str input parameter",
                "required": False,
                },
            REQUIRED_FLD: {
                "order": 5,  # index in the argv list
                "value": "required",
                "type": str,
                "default": "this is a required string",
                "valid_range": None,
                "sflag": "-r",
                "lflag": "--" + REQUIRED_FLD,
                "help": "required input parameter",
                "required": True,
                },
            INPUT_FILE_FLD: {
                "order": 6,  # index in the argv list
                "value": unit_test_argv_file,
                "type": str,
                "default": "this is an input file",
                "valid_range": None,
                "sflag": "-f",
                "lflag": "--" + INPUT_FILE_FLD,
                "help": "input file parameter",
                "required": False,
                },
        }}
        self.valid_sflags = [self.user_inputs[self.user_inputs_parent][k][
            "sflag"] for k in self.user_inputs[self.user_inputs_parent]]


def run_all_tests():
    """
    Run all enabled unit tests.
    """
    # discover all unit test files in current directory
    print("discovering tests...")
    suite = unittest.TestLoader().discover('.', pattern="test_*.py")

    # run all unit tests
    result = unittest.TextTestRunner(verbosity=2).run(suite)

    # flush stdout so that the following output will be at the end
    sys.stdout.flush()
    print("-" * 80)
    print(f"skipped tests({len(result.skipped)}):")
    for name, reason in result.skipped:
        print(name, reason)
    print("-" * 80)

    # set exit code
    if result.wasSuccessful():
        sys.exit(0)
    else:
        sys.exit(1)


def parse_unit_test_runtime_parameters():
    """
    Parse runtime parameters passed in to unit test modules.

    unit test runtime args:
    0 = script_name
    1 = enable integration tests (default = enabled)
    """
    # parameter 1: enable integration tests
    global_par = "ENABLE_FUNCTIONAL_INTEGRATION_TESTS"
    enable_flag = getattr(sys.modules[__name__], global_par)

    # parse runtime parameters
    if len(sys.argv) > 1:
        enable_int_test_flags = ["1", "t", "true"]
        enable_flag = bool(sys.argv[1].lower() in enable_int_test_flags)

    # update global parameter
    setattr(sys.modules[__name__], global_par, enable_flag)
    print(f"integration tests are {['disabled', 'enabled'][enable_flag]}")
    return enable_flag


def mock_exception(exception_type,
                   exception_args):
    """
    Mock an exception.

    inputs:
        exception_type(obj): exception type
        exception_args(list): exception arguments.
    returns:
        None, raises an exception.
    """
    raise exception_type(*exception_args)


def omit_env_vars(target_list):
    """
    Create mock env var dict with specified env vars omitted.

    inputs:
        target_list(ist): env vars to omit.
    returns;
        (dict): env var dict.
    """
    modified_environ = {
        k: v for k, v in os.environ.items() if k not in target_list
    }
    return modified_environ


if __name__ == "__main__":
    parse_unit_test_runtime_parameters()
    print(
        f"DEBUG: ENABLE_FUNCTIONAL_INTEGRATION_TESTS="
        f"{ENABLE_FUNCTIONAL_INTEGRATION_TESTS}")
    print(
        f"DEBUG: ENABLE_PERFORMANCE_INTEGRATION_TESTS="
        f"{ENABLE_PERFORMANCE_INTEGRATION_TESTS}")
    run_all_tests()



"""
Unit test module for utilities.py.
"""
# built-in imports
import os
import shutil
import unittest

# local imports
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class FileAndLoggingTests(utc.UnitTest):
    """Test functions related to logging functions."""

    def setUp(self):
        super().setUp()
        util.log_msg.file_name = "unit_test.txt"

    def test_log_msg_create_folder(self):
        """
        Confirm log_msg() will create folder if needed
        """
        # override data file path
        path_backup = util.FILE_PATH
        util.FILE_PATH = ".//unittest_data"

        file_name = "unit_test.txt"
        full_path = util.get_full_file_path(file_name)
        try:
            # remove directory if it already exists
            if os.path.exists(util.FILE_PATH):
                shutil.rmtree(util.FILE_PATH)

            # write to file and path that does not exist
            test_msg1 = "first test message from unit test"
            # test_msg1_length = util.utf8len(test_msg1 + "\n") + 1
            return_buffer = util.log_msg(test_msg1, mode=util.BOTH_LOG,
                                         file_name=file_name)
            self.assertEqual(return_buffer["status"], util.NO_ERROR)

            # confirm file exists
            file_size_bytes = os.path.getsize(full_path)
            # self.assertEqual(file_size_bytes, test_msg1_length)
            self.assertGreater(file_size_bytes, 30)
        finally:
            # remove the directory
            shutil.rmtree(util.FILE_PATH)
            # restore original data file name
            util.FILE_PATH = path_backup

    def test_log_msg_write(self):
        """
        Confirm log_msg() can write and append to file.
        """
        file_name = "unit_test.txt"
        full_path = util.get_full_file_path(file_name)

        # delete unit test file if it exists
        self.delete_test_file(full_path)

        # write to file that does not exist
        test_msg1 = "first test message from unit test"
        # test_msg1_length = util.utf8len(test_msg1 + "\n") + 1
        return_buffer = util.log_msg(test_msg1, mode=util.BOTH_LOG,
                                     file_name=file_name)
        self.assertEqual(return_buffer["status"], util.NO_ERROR)

        # confirm file exists
        file_size_bytes = os.path.getsize(full_path)
        # self.assertEqual(file_size_bytes, test_msg1_length)
        self.assertGreater(file_size_bytes, 30)

        # append to file that does exist
        test_msg2 = "second test message from unit test"
        # test_msg2_length = util.utf8len(test_msg2 + "\n") + 1
        return_buffer = util.log_msg(test_msg2, mode=util.BOTH_LOG,
                                     file_name=file_name)
        self.assertEqual(return_buffer["status"], util.NO_ERROR)

        # confirm file exists
        file_size_bytes = os.path.getsize(full_path)
        # file size estimate differs per platform, need to refine
        # self.assertEqual(file_size_bytes,
        #                 test_msg1_length + test_msg2_length)
        self.assertGreater(file_size_bytes, 60)

    def test_get_file_size_bytes(self):
        """
        Confirm get_file_size_bytes() works as expected.
        """
        full_path = __file__  # this file

        # assuming file exists, should return non-zero value
        result = util.get_file_size_bytes(full_path)
        self.assertTrue(result > 0,
                        f"file size for existing file is {result}, "
                        f"expected > 0")

        # bogus file, should return zero value
        result = util.get_file_size_bytes("bogus.123")
        self.assertTrue(result == 0,
                        f"file size for bogus file is {result}, expected == 0")

    def test_log_rotate_file(self):
        """
        Confirm log_rotate_file() works as expected.
        """
        file_name = "unit_test.txt"
        full_path = util.get_full_file_path(file_name)
        file_size_bytes = util.get_file_size_bytes(full_path)

        # check while under max limit, should not rotate file
        file_size_bytes_same = util.log_rotate_file(full_path, file_size_bytes,
                                                    file_size_bytes + 1)
        self.assertEqual(file_size_bytes, file_size_bytes_same,
                         f"log_rotate_file under max limit, file size should "
                         f"not change, expected size={file_size_bytes}, actual"
                         f" size={file_size_bytes_same}")

        # check while above max limit, should rotate file and return 0
        file_size_bytes_new = util.log_rotate_file(full_path, file_size_bytes,
                                                   file_size_bytes - 1)
        expected_size = 0
        self.assertEqual(expected_size, file_size_bytes_new,
                         f"log_rotate_file above max limit, file size should "
                         f"be reset to 0, expected size={expected_size}, "
                         f"actual size={file_size_bytes_new}")

    def test_write_to_file(self):
        """
        Verify write_to_file() function.
        """
        file_name = "unit_test.txt"
        full_path = util.get_full_file_path(file_name)

        # delete unit test file if it exists
        self.delete_test_file(full_path)

        # test message
        msg = "unit test bogus message"
        print(f"test message={util.utf8len(msg)} bytes")

        # write to non-existing file, bytes written + EOF == bytes read
        bytes_written = util.write_to_file(full_path, 0, msg)
        bytes_expected = bytes_written + [0, 1][util.is_windows_environment()]
        bytes_present = util.get_file_size_bytes(full_path)
        self.assertEqual(bytes_expected, bytes_present,
                         f"writing to non-existent file, bytes written="
                         f"{bytes_expected}, file size={bytes_present}")

        # write to existing file with reset, bytes written == bytes read
        bytes_written = util.write_to_file(full_path, 0, msg)
        bytes_expected = bytes_written + [0, 1][util.is_windows_environment()]
        bytes_present = util.get_file_size_bytes(full_path)
        self.assertEqual(bytes_expected, bytes_present,
                         f"writing to existing file with override option, "
                         f"bytes written={bytes_expected}, "
                         f"file size={bytes_present}")

        # write to existing file, bytes written < bytes read
        file_size_bytes = util.get_file_size_bytes(full_path)
        bytes_written = util.write_to_file(full_path, file_size_bytes, msg)
        bytes_expected = (bytes_written + file_size_bytes +
                          [0, 1][util.is_windows_environment()])
        bytes_present = util.get_file_size_bytes(full_path)
        self.assertEqual(bytes_expected, bytes_present,
                         f"writing to existent file, bytes "
                         f"expected={bytes_expected}, "
                         f"file size={bytes_present}")

    def test_get_full_file_path(self):
        """
        Verify get_full_file_path() function.
        """
        file_name = "dummy.txt"
        full_path = util.get_full_file_path(file_name)
        expected_value = util.FILE_PATH + "//" + file_name
        print(f"full path={full_path}")
        self.assertEqual(expected_value, full_path,
                         f"expected={expected_value}, actual={full_path}")

    def delete_test_file(self, full_path):
        """Delete the test file.

        inputs:
            full_path(str): full file path
        returns:
            (bool): True if file was deleted, False if it did not exist.
        """
        try:
            os.remove(full_path)
            print(f"unit test file '{full_path}' deleted.")
            return True
        except FileNotFoundError:
            print(f"unit test file '{full_path}' did not exist.")
            return False


class MetricsTests(utc.UnitTest):
    """Test functions related temperature/humidity metrics."""

    def setUp(self):
        super().setUp()
        util.log_msg.file_name = "unit_test.txt"

    def test_temp_value_with_units(self):
        """Verify function attaches units as expected."""

        for test_case in [44, -1, 101, 2, "13", "-13", None]:
            for precision in [0, 1, 2]:
                for disp_unit in ['F', 'c']:
                    print(f"test case: value={test_case}, precision="
                          f"{precision}, units={disp_unit}")
                    if test_case is None:
                        formatted = "None"
                    else:
                        if isinstance(test_case, str):
                            formatted = f"{float(test_case):.{precision}f}"
                        else:
                            formatted = f"{test_case:.{precision}f}"
                    expected_val = f'{formatted}°{disp_unit}'
                    actual_val = util.temp_value_with_units(test_case,
                                                            disp_unit,
                                                            precision)
                    self.assertEqual(expected_val, actual_val,
                                     f"test case: {test_case}, expected_val="
                                     f"{expected_val}, actual_val="
                                     f"{actual_val}")

        # test failing case
        with self.assertRaises(ValueError):
            util.temp_value_with_units(-13, "bogus", 1)

    def test_humidity_value_with_units(self):
        """Verify function attaches units as expected."""

        for test_case in [44, -1, 101, 2, "13", "-13", None]:
            for precision in [0, 1, 2]:
                for disp_unit in ['RH']:
                    print(f"test case: value={test_case}, precision="
                          f"{precision}, units={disp_unit}")
                    if test_case is None:
                        formatted = "None"
                    else:
                        if isinstance(test_case, str):
                            formatted = f"{float(test_case):.{precision}f}"
                        else:
                            formatted = f"{test_case:.{precision}f}"
                    expected_val = f'{formatted}%{disp_unit}'
                    actual_val = util.humidity_value_with_units(test_case,
                                                                disp_unit,
                                                                precision)
                    self.assertEqual(expected_val, actual_val,
                                     f"test case: {test_case}, expected_val="
                                     f"{expected_val}, actual_val="
                                     f"{actual_val}")

        # test failing case
        with self.assertRaises(ValueError):
            util.humidity_value_with_units(-13, "bogus", 1)

    def test_c_to_f(self):
        """Verify C to F calculations."""

        # int and float cases
        for tempc in [0, -19, 34, 101, -44.1, None]:
            tempf = util.c_to_f(tempc)
            if tempc is None:
                expected_tempf = None  # pass-thru
            else:
                expected_tempf = tempc * 9.0 / 5 + 32
            self.assertEqual(expected_tempf, tempf, f"test case {tempc}: "
                             f"expected={expected_tempf}, actual={tempf}")

        # verify exception cases
        for tempc in ['0', "", "*"]:
            with self.assertRaises(TypeError):
                tempf = util.c_to_f(tempc)
            # expected_tempf = tempc
            # self.assertEqual(expected_tempf, tempf, "test case %s: "
            #                  "expected=%s, actual=%s" %
            #                  (tempc, expected_tempf, tempf))

    def test_f_to_c(self):
        """Verify F to C calculations."""

        # int and float cases
        for tempf in [0, -19, 34, 101, -44.1, None]:
            tempc = util.f_to_c(tempf)
            if tempf is None:
                expected_tempc = None  # pass-thru
            else:
                expected_tempc = (tempf - 32) * 5 / 9.0
            self.assertEqual(expected_tempc, tempc, f"test case {tempf}: "
                             f"expected={expected_tempc}, actual={tempc}")

        # verify exception case
        for tempf in ['0', "", "*"]:
            with self.assertRaises(TypeError):
                tempc = util.f_to_c(tempf)
            # expected_tempc = tempf  # pass-thru
            # self.assertEqual(expected_tempc, tempc, "test case %s: "
            #                  "expected=%s, actual=%s" %
            #                  (tempf, expected_tempc, tempc))


class MiscTests(utc.UnitTest):
    """Miscellaneous util tests."""

    def setUp(self):
        super().setUp()
        util.log_msg.file_name = "unit_test.txt"

    def test_get_function_name(self):
        """
        Confirm get_function_name works as expected.
        """
        for test in range(1, 4):
            print(f"get_function_name({test})={util.get_function_name(test)}")

        # default
        test = "<default>"
        print(f"testing util.get_function_name({test})")
        ev_1 = "test_get_function_name"
        result_1 = util.get_function_name()
        print(f"get_function_name({test})={result_1}")
        self.assertEqual(ev_1, result_1, f"expected={ev_1}, actual={result_1}")

        # test 1
        test = 1
        print(f"testing util.get_function_name({test})")
        ev_1 = "test_get_function_name"
        result_1 = util.get_function_name(test)
        print(f"get_function_name({test})={result_1}")
        self.assertEqual(ev_1, result_1, f"test{test}: expected={ev_1}, "
                         f"actual={result_1}")

        # test 2
        test = 2
        print(f"testing util.get_function_name({test})")
        ev_1 = ["patched",  # mock patch decorator
                ]
        result_1 = util.get_function_name(test)
        print(f"get_function_name({test})={result_1}")
        self.assertTrue(result_1 in ev_1, f"test{test}: expected values={ev_1}"
                        f", actual={result_1}")

        # test 3
        test = 3
        print(f"testing util.get_function_name({test})")
        ev_1 = ["run",  # Linux
                "_callTestMethod",  # windows
                ]
        result_1 = util.get_function_name(test)
        print(f"get_function_name({test})={result_1}")
        self.assertTrue(result_1 in ev_1, f"test{test}: expected values={ev_1}"
                        f", actual={result_1}")

    def test_utf8len(self):
        """
        Verify utf8len().
        """
        for test_case in ["A", "BB", "ccc", "dd_d"]:
            print(f"testing util.utf8len({test_case})")
            expected_value = 1 * len(test_case)
            actual_value = util.utf8len(test_case)
            self.assertEqual(expected_value, actual_value,
                             f"expected={expected_value}, "
                             f"actual={actual_value}")

    def test_get_key_from_value(self):
        """Verify get_key_from_value()."""
        test_dict = {'A': 1, 'B': 2, 'C': 1}

        # test keys with distinctvalue, determinant case
        test_case = 2
        expected_val = ['B']
        actual_val = util.get_key_from_value(test_dict, test_case)
        self.assertTrue(actual_val in expected_val,
                        f"test case: {test_case}, expected_val={expected_val},"
                        f" actual_val={actual_val}")

        # test keys with same value, indeterminant case
        test_case = 1
        expected_val = ['A', 'C']
        actual_val = util.get_key_from_value(test_dict, test_case)
        self.assertTrue(actual_val in expected_val,
                        f"test case: {test_case}, expected_val={expected_val},"
                        f" actual_val={actual_val}")

        # test key not found
        with self.assertRaises(KeyError):
            print("attempting to input bad dictionary key, "
                  "expect exception...")
            actual_val = util.get_key_from_value(test_dict, "bogus_value")


class RuntimeParameterTest(utc.RuntimeParameterTest):
    """Generic runtime parameter tests."""

    mod = utc  # module to test
    script = os.path.realpath(__file__)
    bool_val = False
    int_val = 9
    float_val = 8.8
    str_val = "test str"
    required_val = "required"
    input_file = utc.unit_test_argv_file

    # fields for testing, mapped to class variables.
    # (value, field name)
    test_fields = [
        (script, os.path.realpath(__file__)),
        (bool_val, utc.BOOL_FLD),
        (int_val, utc.INT_FLD),
        (float_val, utc.FLOAT_FLD),
        (str_val, utc.STR_FLD),
        (required_val, utc.REQUIRED_FLD),
    ]
    # test case with input file
    test_fields_with_file = [
        (script, os.path.realpath(__file__)),
        (bool_val, utc.BOOL_FLD),
        (int_val, utc.INT_FLD),
        (float_val, utc.FLOAT_FLD),
        (str_val, utc.STR_FLD),
        (required_val, utc.REQUIRED_FLD),
        (input_file, utc.INPUT_FILE_FLD),
        ]


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)
